@@ -62,7 +62,7 @@
 #include <stdint.h>
 
 #include "compositor/weston.h"
-#include "compositor.h"
+#include <libweston/libweston.h>
 #include "ivi-shell.h"
 #include "ivi-layout-export.h"
 #include "ivi-layout-private.h"
@@ -155,8 +155,7 @@ ivi_view_destroy(struct ivi_layout_view *ivi_view)
 
 	if (weston_surface_is_desktop_surface(ivi_view->ivisurf->surface))
 		weston_desktop_surface_unlink_view(ivi_view->view);
-	else
-		weston_view_destroy(ivi_view->view);
+	weston_view_destroy(ivi_view->view);
 
 	free(ivi_view);
 }
@@ -346,9 +345,13 @@ calc_transformation_matrix(struct ivi_rectangle *source_rect,
 	source_center_y = source_rect->y + source_rect->height * 0.5f;
 	weston_matrix_translate(m, -source_center_x, -source_center_y, 0.0f);
 
-	scale_x = (float) dest_rect->width / (float) source_rect->width;
-	scale_y = (float) dest_rect->height / (float) source_rect->height;
-	weston_matrix_scale(m, scale_x, scale_y, 1.0f);
+	if ((dest_rect->width != source_rect->width) ||
+	    (dest_rect->height != source_rect->height))
+	{
+		scale_x = (float) dest_rect->width / (float) source_rect->width;
+		scale_y = (float) dest_rect->height / (float) source_rect->height;
+		weston_matrix_scale(m, scale_x, scale_y, 1.0f);
+	}
 
 	translate_x = dest_rect->width * 0.5f + dest_rect->x;
 	translate_y = dest_rect->height * 0.5f + dest_rect->y;
@@ -699,15 +702,9 @@ commit_surface_list(struct ivi_layout *layout)
 			ivisurf->pending.prop.transition_type = IVI_LAYOUT_TRANSITION_NONE;
 
 			if (configured && !is_surface_transition(ivisurf)) {
-				if (weston_surface_is_desktop_surface(ivisurf->surface)) {
-					weston_desktop_surface_set_size(ivisurf->weston_desktop_surface,
-									ivisurf->prop.dest_width,
-									ivisurf->prop.dest_height);
-				} else {
-					shell_surface_send_configure(ivisurf->surface,
-								     ivisurf->prop.dest_width,
-								     ivisurf->prop.dest_height);
-				}
+				ivi_layout_surface_set_size(ivisurf,
+							    ivisurf->prop.dest_width,
+							    ivisurf->prop.dest_height);
 			}
 		} else {
 			configured = 0;
@@ -721,9 +718,9 @@ commit_surface_list(struct ivi_layout *layout)
 			ivisurf->pending.prop.transition_type = IVI_LAYOUT_TRANSITION_NONE;
 
 			if (configured && !is_surface_transition(ivisurf)) {
-				shell_surface_send_configure(ivisurf->surface,
-							     ivisurf->prop.dest_width,
-							     ivisurf->prop.dest_height);
+				ivi_layout_surface_set_size(ivisurf,
+							    ivisurf->prop.dest_width,
+							    ivisurf->prop.dest_height);
 			}
 		}
 	}
@@ -1556,6 +1553,19 @@ ivi_layout_surface_set_destination_rectangle(struct ivi_layout_surface *ivisurf,
 	return IVI_SUCCEEDED;
 }
 
+void
+ivi_layout_surface_set_size(struct ivi_layout_surface *ivisurf,
+			    int32_t width, int32_t height)
+{
+	if (weston_surface_is_desktop_surface(ivisurf->surface)) {
+		weston_desktop_surface_set_size(ivisurf->weston_desktop_surface,
+						width, height);
+	} else {
+		shell_surface_send_configure(ivisurf->surface,
+					     width, height);
+	}
+}
+
 static int32_t
 ivi_layout_screen_add_layer(struct weston_output *output,
 			    struct ivi_layout_layer *addlayer)
@@ -2043,6 +2053,16 @@ ivi_layout_init_with_compositor(struct weston_compositor *ec)
 				   sizeof(struct ivi_layout_interface));
 }
 
+void
+ivi_layout_fini(void)
+{
+	struct ivi_layout *layout = get_instance();
+
+	weston_layer_fini(&layout->layout_layer);
+
+	/* XXX: tear down everything else */
+}
+
 static struct ivi_layout_interface ivi_layout_interface = {
 	/**
 	 * commit all changes
