@@ -346,11 +346,17 @@ shell_destroy(struct wl_listener *listener, void *data)
 	wl_list_remove(&shell->destroy_listener.link);
 	wl_list_remove(&shell->wake_listener.link);
 
+	weston_desktop_destroy(shell->desktop);
+
 	wl_list_for_each_safe(ivisurf, next, &shell->ivi_surface_list, link) {
+		if (ivisurf->layout_surface != NULL)
+			layout_surface_cleanup(ivisurf);
 		wl_list_remove(&ivisurf->link);
 		free(ivisurf);
 	}
 
+	ivi_layout_fini();
+
 	free(shell);
 }
 
@@ -379,7 +385,7 @@ init_ivi_shell(struct weston_compositor *compositor, struct ivi_shell *shell)
 {
 	struct weston_config *config = wet_get_config(compositor);
 	struct weston_config_section *section;
-	int developermode;
+	bool developermode;
 
 	shell->compositor = compositor;
 
@@ -623,10 +629,14 @@ wet_shell_init(struct weston_compositor *compositor,
 	if (shell == NULL)
 		return -1;
 
-	init_ivi_shell(compositor, shell);
+	if (!weston_compositor_add_destroy_listener_once(compositor,
+							 &shell->destroy_listener,
+							 shell_destroy)) {
+		free(shell);
+		return 0;
+	}
 
-	shell->destroy_listener.notify = shell_destroy;
-	wl_signal_add(&compositor->destroy_signal, &shell->destroy_listener);
+	init_ivi_shell(compositor, shell);
 
 	shell->wake_listener.notify = wake_handler;
 	wl_signal_add(&compositor->wake_signal, &shell->wake_listener);
@@ -649,6 +659,7 @@ err_desktop:
 	weston_desktop_destroy(shell->desktop);
 
 err_shell:
+	wl_list_remove(&shell->destroy_listener.link);
 	free(shell);
 
 	return IVI_FAILED;
