@@ -32,7 +32,8 @@
 #include <stdio.h>
 #include <assert.h>
 
-#include "compositor.h"
+#include <libweston/libweston.h>
+#include "libweston-internal.h"
 #include "shared/helpers.h"
 #include "shared/timespec-util.h"
 
@@ -220,6 +221,13 @@ data_offer_finish(struct wl_client *client, struct wl_resource *resource)
 	if (!offer->source || offer->source->offer != offer)
 		return;
 
+	if (offer->source->set_selection) {
+		wl_resource_post_error(offer->resource,
+				       WL_DATA_OFFER_ERROR_INVALID_FINISH,
+				       "finish only valid for drag n drop");
+		return;
+	}
+
 	/* Disallow finish while we have a grab driving drag-and-drop, or
 	 * if the negotiation is not at the right stage
 	 */
@@ -393,7 +401,7 @@ data_source_set_actions(struct wl_client *client,
 	source->actions_set = true;
 }
 
-static struct wl_data_source_interface data_source_interface = {
+static const struct wl_data_source_interface data_source_interface = {
 	data_source_offer,
 	data_source_destroy,
 	data_source_set_actions
@@ -632,7 +640,8 @@ data_device_end_pointer_drag_grab(struct weston_pointer_drag *drag)
 
 	data_device_end_drag_grab(&drag->base, pointer->seat);
 	weston_pointer_end_grab(pointer);
-	weston_keyboard_end_grab(keyboard);
+	if (keyboard)
+		weston_keyboard_end_grab(keyboard);
 	free(drag);
 }
 
@@ -732,7 +741,8 @@ data_device_end_touch_drag_grab(struct weston_touch_drag *drag)
 
 	data_device_end_drag_grab(&drag->base, touch->seat);
 	weston_touch_end_grab(touch);
-	weston_keyboard_end_grab(keyboard);
+	if (keyboard)
+		weston_keyboard_end_grab(keyboard);
 	free(drag);
 }
 
@@ -749,8 +759,10 @@ drag_grab_touch_up(struct weston_touch_grab *grab,
 
 	if (touch_drag->base.focus_resource)
 		wl_data_device_send_drop(touch_drag->base.focus_resource);
-	if (touch_drag->base.data_source)
+	if (touch_drag->base.data_source) {
+		touch_drag->base.data_source->seat = NULL;
 		wl_list_remove(&touch_drag->base.data_source_listener.link);
+	}
 	data_device_end_touch_drag_grab(touch_drag);
 }
 
@@ -944,10 +956,12 @@ weston_pointer_start_drag(struct weston_pointer *pointer,
 	}
 
 	weston_pointer_clear_focus(pointer);
-	weston_keyboard_set_focus(keyboard, NULL);
+	if (keyboard)
+		weston_keyboard_set_focus(keyboard, NULL);
 
 	weston_pointer_start_grab(pointer, &drag->grab);
-	weston_keyboard_start_grab(keyboard, &drag->base.keyboard_grab);
+	if (keyboard)
+		weston_keyboard_start_grab(keyboard, &drag->base.keyboard_grab);
 
 	return 0;
 }
@@ -1004,10 +1018,12 @@ weston_touch_start_drag(struct weston_touch *touch,
 			      &drag->base.data_source_listener);
 	}
 
-	weston_keyboard_set_focus(keyboard, NULL);
+	if (keyboard)
+		weston_keyboard_set_focus(keyboard, NULL);
 
 	weston_touch_start_grab(touch, &drag->grab);
-	weston_keyboard_start_grab(keyboard, &drag->base.keyboard_grab);
+	if (keyboard)
+		weston_keyboard_start_grab(keyboard, &drag->base.keyboard_grab);
 
 	drag_grab_touch_focus(drag);
 
@@ -1146,6 +1162,9 @@ weston_seat_set_selection(struct weston_seat *seat,
 	seat->selection_data_source = source;
 	seat->selection_serial = serial;
 
+	if (source)
+		source->set_selection = true;
+
 	if (keyboard)
 		focus = keyboard->focus;
 	if (focus && focus->resource) {
@@ -1267,6 +1286,7 @@ create_data_source(struct wl_client *client,
 	source->dnd_actions = 0;
 	source->current_dnd_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
 	source->compositor_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
+	source->set_selection = false;
 
 	wl_array_init(&source->mime_types);
 
