@@ -38,7 +38,8 @@
 #include <systemd/sd-login.h>
 #include <unistd.h>
 
-#include "compositor.h"
+#include <libweston/libweston.h>
+#include "backend.h"
 #include "dbus.h"
 #include "launcher-impl.h"
 
@@ -96,6 +97,7 @@ launcher_logind_take_device(struct launcher_logind *wl, uint32_t major,
 	reply = dbus_connection_send_with_reply_and_block(wl->dbus, m,
 							  -1, NULL);
 	if (!reply) {
+		weston_log("logind: TakeDevice on %d:%d failed.\n", major, minor);
 		r = -ENODEV;
 		goto err_unref;
 	}
@@ -105,6 +107,7 @@ launcher_logind_take_device(struct launcher_logind *wl, uint32_t major,
 				  DBUS_TYPE_BOOLEAN, &paused,
 				  DBUS_TYPE_INVALID);
 	if (!b) {
+		weston_log("logind: error parsing reply to TakeDevice.\n");
 		r = -ENODEV;
 		goto err_reply;
 	}
@@ -172,17 +175,25 @@ launcher_logind_open(struct weston_launcher *launcher, const char *path, int fla
 	int fl, r, fd;
 
 	r = stat(path, &st);
-	if (r < 0)
+	if (r < 0) {
+		weston_log("logind: cannot stat: %s! error=%s\n", path, strerror(errno));
 		return -1;
+	}
+
 	if (!S_ISCHR(st.st_mode)) {
+		weston_log("logind: %s is not a character special file!\n", path);
 		errno = ENODEV;
 		return -1;
 	}
 
 	fd = launcher_logind_take_device(wl, major(st.st_rdev),
 				       minor(st.st_rdev), NULL);
-	if (fd < 0)
-		return fd;
+	if (fd < 0) {
+		weston_log("logind: TakeDevice on %s failed, error=%s\n",
+			   path, strerror(-fd));
+		errno = -fd;
+		return -1;
+	}
 
 	/* Compared to weston_launcher_open() we cannot specify the open-mode
 	 * directly. Instead, logind passes us an fd with sane default modes.
@@ -194,6 +205,7 @@ launcher_logind_open(struct weston_launcher *launcher, const char *path, int fla
 	fl = fcntl(fd, F_GETFL);
 	if (fl < 0) {
 		r = -errno;
+		weston_log("logind: cannot get file flags: %s\n", strerror(errno));
 		goto err_close;
 	}
 
@@ -203,6 +215,7 @@ launcher_logind_open(struct weston_launcher *launcher, const char *path, int fla
 	r = fcntl(fd, F_SETFL, fl);
 	if (r < 0) {
 		r = -errno;
+		weston_log("logind: cannot set O_NONBLOCK: %s\n", strerror(errno));
 		goto err_close;
 	}
 	return fd;
@@ -225,7 +238,7 @@ launcher_logind_close(struct weston_launcher *launcher, int fd)
 	r = fstat(fd, &st);
 	close(fd);
 	if (r < 0) {
-		weston_log("logind: cannot fstat fd: %m\n");
+		weston_log("logind: cannot fstat fd: %s\n", strerror(errno));
 		return;
 	}
 
@@ -272,7 +285,7 @@ launcher_logind_activate_vt(struct weston_launcher *launcher, int vt)
 static void
 launcher_logind_set_active(struct launcher_logind *wl, bool active)
 {
-	if (!wl->compositor->session_active == !active)
+	if (wl->compositor->session_active == active)
 		return;
 
 	wl->compositor->session_active = active;
@@ -488,20 +501,21 @@ device_paused(struct launcher_logind *wl, DBusMessage *m)
 	if (!strcmp(type, "pause"))
 		launcher_logind_pause_device_complete(wl, major, minor);
 
-	if (wl->sync_drm && major == DRM_MAJOR)
-		launcher_logind_set_active(wl, false);
+	if (wl->sync_drm && wl->compositor->backend->device_changed)
+		wl->compositor->backend->device_changed(wl->compositor,
+							makedev(major,minor),
+							false);
 }
 
 static void
 device_resumed(struct launcher_logind *wl, DBusMessage *m)
 {
 	bool r;
-	uint32_t major;
+	uint32_t major, minor;
 
 	r = dbus_message_get_args(m, NULL,
 				  DBUS_TYPE_UINT32, &major,
-				  /*DBUS_TYPE_UINT32, &minor,
-				  DBUS_TYPE_UNIX_FD, &fd,*/
+				  DBUS_TYPE_UINT32, &minor,
 				  DBUS_TYPE_INVALID);
 	if (!r) {
 		weston_log("logind: cannot parse ResumeDevice dbus signal\n");
@@ -514,8 +528,10 @@ device_resumed(struct launcher_logind *wl, DBusMessage *m)
 	 * there is no need for us to handle this event for evdev. For DRM, we
 	 * notify the compositor to wake up. */
 
-	if (wl->sync_drm && major == DRM_MAJOR)
-		launcher_logind_set_active(wl, true);
+	if (wl->sync_drm && wl->compositor->backend->device_changed)
+		wl->compositor->backend->device_changed(wl->compositor,
+							makedev(major,minor),
+							true);
 }
 
 static DBusHandlerResult
@@ -718,6 +734,29 @@ launcher_logind_activate(struct launcher_logind *wl)
 	return 0;
 }
 
+static int
+launcher_logind_get_session(char **session)
+{
+	int r;
+
+	r = sd_pid_get_session(getpid(), session);
+	if (r < 0) {
+		if (r != -ENODATA) {
+			weston_log("logind: not running in a systemd session: %d\n", r);
+			return r;
+		}
+	} else {
+		return r;
+	}
+
+	/* When not inside a systemd session look if there is a suitable one */
+	r = sd_uid_get_display(getuid(), session);
+	if (r < 0)
+		weston_log("logind: cannot find systemd session for uid: %d %d\n", getuid(), r);
+
+	return r;
+}
+
 static int
 launcher_logind_connect(struct weston_launcher **out, struct weston_compositor *compositor,
 			int tty, const char *seat_id, bool sync_drm)
@@ -743,11 +782,9 @@ launcher_logind_connect(struct weston_launcher **out, struct weston_compositor *
 		goto err_wl;
 	}
 
-	r = sd_pid_get_session(getpid(), &wl->sid);
-	if (r < 0) {
-		weston_log("logind: not running in a systemd session\n");
+	r = launcher_logind_get_session(&wl->sid);
+	if (r < 0)
 		goto err_seat;
-	}
 
 	t = NULL;
 	r = sd_session_get_seat(wl->sid, &t);
@@ -763,9 +800,9 @@ launcher_logind_connect(struct weston_launcher **out, struct weston_compositor *
 		goto err_session;
 	}
 
-	r = strcmp(t, "seat0");
+	r = sd_seat_can_tty(t);
 	free(t);
-	if (r == 0) {
+	if (r > 0) {
 		r = weston_sd_session_get_vt(wl->sid, &wl->vtnr);
 		if (r < 0) {
 			weston_log("logind: session not running on a VT\n");
@@ -776,6 +813,10 @@ launcher_logind_connect(struct weston_launcher **out, struct weston_compositor *
 			r = -EINVAL;
 			goto err_session;
 		}
+	} else if (r < 0) {
+		weston_log("logind: could not determine if seat %s has ttys or not", t);
+		r = -EINVAL;
+		goto err_session;
 	}
 
 	loop = wl_display_get_event_loop(compositor->wl_display);
@@ -812,7 +853,8 @@ err_seat:
 err_wl:
 	free(wl);
 err_out:
-	weston_log("logind: cannot setup systemd-logind helper (%d), using legacy fallback\n", r);
+	weston_log("logind: cannot setup systemd-logind helper error: (%s), using legacy fallback\n",
+		   strerror(-r));
 	errno = -r;
 	return -1;
 }
@@ -839,10 +881,14 @@ static int
 launcher_logind_get_vt(struct weston_launcher *launcher)
 {
 	struct launcher_logind *wl = wl_container_of(launcher, wl, base);
+	if (wl->vtnr <= 0) {
+		return -EINVAL;
+	}
 	return wl->vtnr;
 }
 
 const struct launcher_interface launcher_logind_iface = {
+	.name = "logind",
 	.connect = launcher_logind_connect,
 	.destroy = launcher_logind_destroy,
 	.open = launcher_logind_open,
