@@ -46,6 +46,7 @@
 struct screenshooter_frame_listener {
 	struct wl_listener listener;
 	struct weston_buffer *buffer;
+	struct weston_output *output;
 	weston_screenshooter_done_func_t done;
 	void *data;
 };
@@ -119,12 +120,12 @@ screenshooter_frame_notify(struct wl_listener *listener, void *data)
 	struct screenshooter_frame_listener *l =
 		container_of(listener,
 			     struct screenshooter_frame_listener, listener);
-	struct weston_output *output = data;
+	struct weston_output *output = l->output;
 	struct weston_compositor *compositor = output->compositor;
 	int32_t stride;
 	uint8_t *pixels, *d, *s;
 
-	output->disable_planes--;
+	weston_output_disable_planes_decr(output);
 	wl_list_remove(&listener->link);
 	stride = l->buffer->width * (PIXMAN_FORMAT_BPP(compositor->read_format) / 8);
 	pixels = malloc(stride * l->buffer->height);
@@ -202,12 +203,13 @@ weston_screenshooter_shoot(struct weston_output *output,
 	}
 
 	l->buffer = buffer;
+	l->output = output;
 	l->done = done;
 	l->data = data;
 	l->listener.notify = screenshooter_frame_notify;
 	wl_signal_add(&output->frame_signal, &l->listener);
-	output->disable_planes++;
-	weston_output_damage(output);
+	weston_output_disable_planes_incr(output);
+	weston_output_schedule_repaint(output);
 
 	return 0;
 }
@@ -261,7 +263,7 @@ weston_recorder_frame_notify(struct wl_listener *listener, void *data)
 {
 	struct weston_recorder *recorder =
 		container_of(listener, struct weston_recorder, frame_listener);
-	struct weston_output *output = data;
+	struct weston_output *output = recorder->output;
 	struct weston_compositor *compositor = output->compositor;
 	uint32_t msecs = timespec_to_msec(&output->frame_time);
 	pixman_box32_t *r;
@@ -285,8 +287,7 @@ weston_recorder_frame_notify(struct wl_listener *listener, void *data)
 
 	pixman_region32_init(&damage);
 	pixman_region32_init(&transformed_damage);
-	pixman_region32_intersect(&damage, &output->region,
-				  &output->previous_damage);
+	pixman_region32_intersect(&damage, &output->region, data);
 	pixman_region32_translate(&damage, -output->x, -output->y);
 	weston_transformed_region(output->width, output->height,
 				 output->transform, output->current_scale,
@@ -445,7 +446,7 @@ weston_recorder_create(struct weston_output *output, const char *filename)
 
 	recorder->frame_listener.notify = weston_recorder_frame_notify;
 	wl_signal_add(&output->frame_signal, &recorder->frame_listener);
-	output->disable_planes++;
+	weston_output_disable_planes_incr(output);
 	weston_output_damage(output);
 
 	return recorder;
@@ -460,7 +461,7 @@ weston_recorder_destroy(struct weston_recorder *recorder)
 {
 	wl_list_remove(&recorder->frame_listener.link);
 	close(recorder->fd);
-	recorder->output->disable_planes--;
+	weston_output_disable_planes_decr(recorder->output);
 	weston_recorder_free(recorder);
 }
 
