@@ -29,6 +29,20 @@
 #include <string.h>
 
 #include "weston-test-client-helper.h"
+#include "weston-test-fixture-compositor.h"
+
+#define VERBOSE 0
+
+static enum test_result_code
+fixture_setup(struct weston_test_harness *harness)
+{
+	struct compositor_setup setup;
+
+	compositor_setup_defaults(&setup);
+
+	return weston_test_harness_execute_as_client(harness, &setup);
+}
+DECLARE_FIXTURE_SETUP(fixture_setup);
 
 #define NUM_SUBSURFACES 3
 
@@ -86,6 +100,22 @@ populate_compound_surface(struct compound_surface *com, struct client *client)
 	}
 }
 
+static void
+fini_compound_surface(struct compound_surface *com)
+{
+	int i;
+
+	for (i = 0; i < NUM_SUBSURFACES; i++) {
+		if (com->sub[i])
+			wl_subsurface_destroy(com->sub[i]);
+		if (com->child[i])
+			wl_surface_destroy(com->child[i]);
+	}
+
+	wl_surface_destroy(com->parent);
+	wl_subcompositor_destroy(com->subco);
+}
+
 TEST(test_subsurface_basic_protocol)
 {
 	struct client *client;
@@ -99,6 +129,10 @@ TEST(test_subsurface_basic_protocol)
 	populate_compound_surface(&com2, client);
 
 	client_roundtrip(client);
+
+	fini_compound_surface(&com1);
+	fini_compound_surface(&com2);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_position_protocol)
@@ -116,6 +150,9 @@ TEST(test_subsurface_position_protocol)
 					   (i + 2) * 20, (i + 2) * 10);
 
 	client_roundtrip(client);
+
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_placement_protocol)
@@ -134,6 +171,9 @@ TEST(test_subsurface_placement_protocol)
 	wl_subsurface_place_below(com.sub[1], com.parent);
 
 	client_roundtrip(client);
+
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_paradox)
@@ -141,6 +181,7 @@ TEST(test_subsurface_paradox)
 	struct client *client;
 	struct wl_surface *parent;
 	struct wl_subcompositor *subco;
+	struct wl_subsurface *sub;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -149,16 +190,22 @@ TEST(test_subsurface_paradox)
 	parent = wl_compositor_create_surface(client->wl_compositor);
 
 	/* surface is its own parent */
-	wl_subcompositor_get_subsurface(subco, parent, parent);
+	sub = wl_subcompositor_get_subsurface(subco, parent, parent);
 
 	expect_protocol_error(client, &wl_subcompositor_interface,
 			      WL_SUBCOMPOSITOR_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(parent);
+	wl_subcompositor_destroy(subco);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_identical_link)
 {
 	struct client *client;
 	struct compound_surface com;
+	struct wl_subsurface *sub;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -166,10 +213,14 @@ TEST(test_subsurface_identical_link)
 	populate_compound_surface(&com, client);
 
 	/* surface is already a subsurface */
-	wl_subcompositor_get_subsurface(com.subco, com.child[0], com.parent);
+	sub = wl_subcompositor_get_subsurface(com.subco, com.child[0], com.parent);
 
 	expect_protocol_error(client, &wl_subcompositor_interface,
 			      WL_SUBCOMPOSITOR_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_change_link)
@@ -177,6 +228,7 @@ TEST(test_subsurface_change_link)
 	struct client *client;
 	struct compound_surface com;
 	struct wl_surface *stranger;
+	struct wl_subsurface *sub;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -185,10 +237,15 @@ TEST(test_subsurface_change_link)
 	populate_compound_surface(&com, client);
 
 	/* surface is already a subsurface */
-	wl_subcompositor_get_subsurface(com.subco, com.child[0], stranger);
+	sub = wl_subcompositor_get_subsurface(com.subco, com.child[0], stranger);
 
 	expect_protocol_error(client, &wl_subcompositor_interface,
 			      WL_SUBCOMPOSITOR_ERROR_BAD_SURFACE);
+
+	fini_compound_surface(&com);
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(stranger);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_nesting)
@@ -196,6 +253,7 @@ TEST(test_subsurface_nesting)
 	struct client *client;
 	struct compound_surface com;
 	struct wl_surface *stranger;
+	struct wl_subsurface *sub;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -204,9 +262,14 @@ TEST(test_subsurface_nesting)
 	populate_compound_surface(&com, client);
 
 	/* parent is a sub-surface */
-	wl_subcompositor_get_subsurface(com.subco, stranger, com.child[0]);
+	sub = wl_subcompositor_get_subsurface(com.subco, stranger, com.child[0]);
 
 	client_roundtrip(client);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(stranger);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_nesting_parent)
@@ -214,6 +277,7 @@ TEST(test_subsurface_nesting_parent)
 	struct client *client;
 	struct compound_surface com;
 	struct wl_surface *stranger;
+	struct wl_subsurface *sub;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -222,16 +286,23 @@ TEST(test_subsurface_nesting_parent)
 	populate_compound_surface(&com, client);
 
 	/* surface is already a parent */
-	wl_subcompositor_get_subsurface(com.subco, com.parent, stranger);
+	sub = wl_subcompositor_get_subsurface(com.subco, com.parent, stranger);
 
 	client_roundtrip(client);
+
+	wl_subsurface_destroy(sub);
+	fini_compound_surface(&com);
+	wl_surface_destroy(stranger);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_loop_paradox)
 {
 	struct client *client;
 	struct wl_surface *surface[3];
+	struct wl_subsurface *sub[3];
 	struct wl_subcompositor *subco;
+	unsigned i;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -242,12 +313,20 @@ TEST(test_subsurface_loop_paradox)
 	surface[2] = wl_compositor_create_surface(client->wl_compositor);
 
 	/* create a nesting loop */
-	wl_subcompositor_get_subsurface(subco, surface[1], surface[0]);
-	wl_subcompositor_get_subsurface(subco, surface[2], surface[1]);
-	wl_subcompositor_get_subsurface(subco, surface[0], surface[2]);
+	sub[0] = wl_subcompositor_get_subsurface(subco, surface[1], surface[0]);
+	sub[1] = wl_subcompositor_get_subsurface(subco, surface[2], surface[1]);
+	sub[2] = wl_subcompositor_get_subsurface(subco, surface[0], surface[2]);
 
 	expect_protocol_error(client, &wl_subcompositor_interface,
 			      WL_SUBCOMPOSITOR_ERROR_BAD_SURFACE);
+
+	for (i = 0; i < ARRAY_LENGTH(sub); i++) {
+		wl_subsurface_destroy(sub[i]);
+		wl_surface_destroy(surface[i]);
+	}
+
+	wl_subcompositor_destroy(subco);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_above_nested_parent)
@@ -270,6 +349,12 @@ TEST(test_subsurface_place_above_nested_parent)
 	wl_subsurface_place_above(sub, com.child[0]);
 
 	client_roundtrip(client);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_above_grandparent)
@@ -294,6 +379,12 @@ TEST(test_subsurface_place_above_grandparent)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_above_great_aunt)
@@ -318,6 +409,12 @@ TEST(test_subsurface_place_above_great_aunt)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_above_child)
@@ -326,6 +423,7 @@ TEST(test_subsurface_place_above_child)
 	struct compound_surface com;
 	struct wl_surface *grandchild;
 	struct wl_subcompositor *subco;
+	struct wl_subsurface *sub;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -334,13 +432,19 @@ TEST(test_subsurface_place_above_child)
 
 	subco = get_subcompositor(client);
 	grandchild = wl_compositor_create_surface(client->wl_compositor);
-	wl_subcompositor_get_subsurface(subco, grandchild, com.child[0]);
+	sub = wl_subcompositor_get_subsurface(subco, grandchild, com.child[0]);
 
 	/* can't place a subsurface above its own child subsurface */
 	wl_subsurface_place_above(com.sub[0], grandchild);
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_below_nested_parent)
@@ -363,6 +467,12 @@ TEST(test_subsurface_place_below_nested_parent)
 	wl_subsurface_place_below(sub, com.child[0]);
 
 	client_roundtrip(client);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_below_grandparent)
@@ -387,6 +497,12 @@ TEST(test_subsurface_place_below_grandparent)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_below_great_aunt)
@@ -411,6 +527,12 @@ TEST(test_subsurface_place_below_great_aunt)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_below_child)
@@ -419,6 +541,7 @@ TEST(test_subsurface_place_below_child)
 	struct compound_surface com;
 	struct wl_surface *grandchild;
 	struct wl_subcompositor *subco;
+	struct wl_subsurface *sub;
 
 	client = create_client_and_test_surface(100, 50, 123, 77);
 	assert(client);
@@ -427,13 +550,19 @@ TEST(test_subsurface_place_below_child)
 
 	subco = get_subcompositor(client);
 	grandchild = wl_compositor_create_surface(client->wl_compositor);
-	wl_subcompositor_get_subsurface(subco, grandchild, com.child[0]);
+	sub = wl_subcompositor_get_subsurface(subco, grandchild, com.child[0]);
 
 	/* can't place a subsurface below its own child subsurface */
 	wl_subsurface_place_below(com.sub[0], grandchild);
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_subsurface_destroy(sub);
+	wl_surface_destroy(grandchild);
+	wl_subcompositor_destroy(subco);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_above_stranger)
@@ -453,6 +582,10 @@ TEST(test_subsurface_place_above_stranger)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_surface_destroy(stranger);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_below_stranger)
@@ -472,6 +605,10 @@ TEST(test_subsurface_place_below_stranger)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	wl_surface_destroy(stranger);
+	fini_compound_surface(&com);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_above_foreign)
@@ -491,6 +628,10 @@ TEST(test_subsurface_place_above_foreign)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	fini_compound_surface(&com1);
+	fini_compound_surface(&com2);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_place_below_foreign)
@@ -510,6 +651,10 @@ TEST(test_subsurface_place_below_foreign)
 
 	expect_protocol_error(client, &wl_subsurface_interface,
 			      WL_SUBSURFACE_ERROR_BAD_SURFACE);
+
+	fini_compound_surface(&com1);
+	fini_compound_surface(&com2);
+	client_destroy(client);
 }
 
 TEST(test_subsurface_destroy_protocol)
@@ -543,6 +688,8 @@ TEST(test_subsurface_destroy_protocol)
 	wl_subsurface_destroy(com.sub[1]);
 
 	client_roundtrip(client);
+
+	client_destroy(client);
 }
 
 static void
@@ -609,6 +756,8 @@ create_subsurface_tree(struct client *client, struct wl_surface **surfs,
 
 #undef SUB_LINK
 	};
+
+	wl_subcompositor_destroy(subco);
 }
 
 static void
@@ -694,11 +843,13 @@ destroy_permu_object(struct wl_surface **surfs,
 	int h = (i + 1) / 2;
 
 	if (i & 1) {
-		testlog(" [sub  %2d]", h);
+		if (VERBOSE)
+			testlog(" [sub  %2d]", h);
 		wl_subsurface_destroy(subs[h]);
 		subs[h] = NULL;
 	} else {
-		testlog(" [surf %2d]", h);
+		if (VERBOSE)
+			testlog(" [surf %2d]", h);
 		wl_surface_destroy(surfs[h]);
 		surfs[h] = NULL;
 	}
@@ -736,15 +887,18 @@ TEST(test_subsurface_destroy_permutations)
 
 		create_subsurface_tree(client, surfs, subs, test_size);
 
-		testlog("permu");
+		if (VERBOSE)
+			testlog("permu");
 
-		for (i = 0; i < NSTEPS; i++)
-			testlog(" %2d", per.cnt[i]);
+		if (VERBOSE)
+			for (i = 0; i < NSTEPS; i++)
+				testlog(" %2d", per.cnt[i]);
 
 		for (i = 0; i < NSTEPS; i++)
 			destroy_permu_object(surfs, subs, per.cnt[i]);
 
-		testlog("\n");
+		if (VERBOSE)
+			testlog("\n");
 		client_roundtrip(client);
 
 		destroy_subsurface_tree(surfs, subs, test_size);
@@ -753,4 +907,6 @@ TEST(test_subsurface_destroy_permutations)
 
 	client_roundtrip(client);
 	testlog("tried %d destroy permutations\n", counter);
+
+	client_destroy(client);
 }
