@@ -37,9 +37,18 @@
 #include <wayland-client-protocol.h>
 #include "weston-test-runner.h"
 #include "weston-test-client-protocol.h"
+#include "weston-screenshooter-client-protocol.h"
+#include "viewporter-client-protocol.h"
 
 struct client {
 	struct wl_display *wl_display;
+
+	/*
+	 * Have successfully received an expected protocol error, the
+	 * connection is in error state, and that is ok.
+	 */
+	bool errored_ok;
+
 	struct wl_registry *wl_registry;
 	struct wl_compositor *wl_compositor;
 	struct wl_shm *wl_shm;
@@ -56,7 +65,9 @@ struct client {
 	struct surface *surface;
 	int has_argb;
 	struct wl_list global_list;
-	bool has_wl_drm;
+	struct wl_list output_list; /* struct output::link */
+	struct weston_screenshooter *screenshooter;
+	bool buffer_copy_done;
 };
 
 struct global {
@@ -71,7 +82,6 @@ struct test {
 	int pointer_x;
 	int pointer_y;
 	uint32_t n_egl_buffers;
-	int buffer_copy_done;
 };
 
 struct input {
@@ -89,6 +99,7 @@ struct input {
 struct pointer {
 	struct wl_pointer *wl_pointer;
 	struct surface *focus;
+	uint32_t serial;
 	int x;
 	int y;
 	uint32_t button;
@@ -145,6 +156,7 @@ struct touch {
 
 struct output {
 	struct wl_output *wl_output;
+	struct wl_list link; /* struct client::output_list */
 	int x;
 	int y;
 	int width;
@@ -161,7 +173,7 @@ struct buffer {
 
 struct surface {
 	struct wl_surface *wl_surface;
-	struct output *output;
+	struct output *output; /* not owned */
 	int x;
 	int y;
 	int width;
@@ -176,12 +188,23 @@ struct rectangle {
 	int height;
 };
 
+struct range {
+	int a;
+	int b;
+};
+
 struct client *
 create_client(void);
 
+void
+client_destroy(struct client *client);
+
 struct surface *
 create_test_surface(struct client *client);
 
+void
+surface_destroy(struct surface *surface);
+
 struct client *
 create_client_and_test_surface(int x, int y, int width, int height);
 
@@ -222,13 +245,18 @@ screenshot_output_filename(const char *basename, uint32_t seq);
 char *
 screenshot_reference_filename(const char *basename, uint32_t seq);
 
+char *
+image_filename(const char *basename);
+
 bool
 check_images_match(pixman_image_t *img_a, pixman_image_t *img_b,
-		   const struct rectangle *clip);
+		   const struct rectangle *clip,
+		   const struct range *prec);
 
 pixman_image_t *
 visualize_image_difference(pixman_image_t *img_a, pixman_image_t *img_b,
-			   const struct rectangle *clip_rect);
+			   const struct rectangle *clip_rect,
+			   const struct range *prec);
 
 bool
 write_image_as_png(pixman_image_t *image, const char *fname);
@@ -239,4 +267,37 @@ load_image_from_png(const char *fname);
 struct buffer *
 capture_screenshot_of_output(struct client *client);
 
+bool
+verify_image(struct buffer *shot,
+	     const char *ref_image,
+	     int ref_seq_no,
+	     const struct rectangle *clip,
+	     int seq_no);
+
+bool
+verify_screen_content(struct client *client,
+		      const char *ref_image,
+		      int ref_seq_no,
+		      const struct rectangle *clip,
+		      int seq_no);
+
+struct buffer *
+client_buffer_from_image_file(struct client *client,
+			      const char *basename,
+			      int scale);
+
+void *
+bind_to_singleton_global(struct client *client,
+			 const struct wl_interface *iface,
+			 int version);
+
+struct wp_viewport *
+client_create_viewport(struct client *client);
+
+void
+fill_image_with_color(pixman_image_t *image, const pixman_color_t *color);
+
+pixman_color_t *
+color_rgb888(pixman_color_t *tmp, uint8_t r, uint8_t g, uint8_t b);
+
 #endif
