@@ -42,6 +42,19 @@
 #include <string.h>
 
 #include "weston-test-runner.h"
+#include "weston-test-fixture-compositor.h"
+
+static enum test_result_code
+fixture_setup(struct weston_test_harness *harness)
+{
+	struct compositor_setup setup;
+
+	compositor_setup_defaults(&setup);
+	setup.xwayland = true;
+
+	return weston_test_harness_execute_as_client(harness, &setup);
+}
+DECLARE_FIXTURE_SETUP(fixture_setup);
 
 TEST(xwayland_client_test)
 {
@@ -53,6 +66,9 @@ TEST(xwayland_client_test)
 	Atom atom, type_atom, actual_type;
 	char *wm_name;
 
+	if (access(XSERVER_PATH, X_OK) != 0)
+		exit(77);
+
 	display = XOpenDisplay(NULL);
 	if (!display)
 		exit(EXIT_FAILURE);
@@ -92,7 +108,6 @@ TEST(xwayland_client_test)
 	XSelectInput(display, window, ExposureMask);
 	XMapWindow(display, window);
 
-	alarm(4);
 	while (1) {
 		XNextEvent(display, &event);
 		if (event.type == Expose)
@@ -100,5 +115,4 @@ TEST(xwayland_client_test)
 	}
 
 	XCloseDisplay(display);
-	exit(EXIT_SUCCESS);
 }
